@@ -14,6 +14,7 @@ export
     getnameinfo,
     getipaddr,
     getipaddrs,
+    islinklocaladdr,
     getpeername,
     getsockname,
     listen,
@@ -21,6 +22,8 @@ export
     recv,
     recvfrom,
     send,
+    join_multicast_group,
+    leave_multicast_group,
     TCPSocket,
     UDPSocket,
     @ip_str,
@@ -231,7 +234,7 @@ const UV_UDP_REUSEADDR = 4
 
 ##
 
-function _bind(sock::TCPServer, host::Union{IPv4, IPv6}, port::UInt16, flags::UInt32=UInt32(0))
+function _bind(sock::Union{TCPServer, TCPSocket}, host::Union{IPv4, IPv6}, port::UInt16, flags::UInt32=UInt32(0))
     host_in = Ref(hton(host.host))
     return ccall(:jl_tcp_bind, Int32, (Ptr{Cvoid}, UInt16, Ptr{Cvoid}, Cuint, Cint),
             sock, hton(port), host_in, flags, host isa IPv6)
@@ -252,7 +255,7 @@ Bind `socket` to the given `host:port`. Note that `0.0.0.0` will listen on all d
 * If `reuseaddr=true`, multiple threads or processes can bind to the same address without error
   if they all set `reuseaddr=true`, but only the last to bind will receive any traffic.
 """
-function bind(sock::Union{TCPServer, UDPSocket}, host::IPAddr, port::Integer; ipv6only = false, reuseaddr = false, kws...)
+function bind(sock::Union{TCPServer, UDPSocket, TCPSocket}, host::IPAddr, port::Integer; ipv6only = false, reuseaddr = false, kws...)
     if sock.status != StatusInit
         error("$(typeof(sock)) is not in initialization state")
     end
@@ -274,7 +277,9 @@ function bind(sock::Union{TCPServer, UDPSocket}, host::IPAddr, port::Integer; ip
             return false
         end
     end
-    sock.status = StatusOpen
+    if isa(sock, TCPServer) || isa(sock, UDPSocket)
+        sock.status = StatusOpen
+    end
     isa(sock, UDPSocket) && setopt(sock; kws...)
     iolock_end()
     return true
@@ -724,6 +729,56 @@ end
 
 listenany(default_port) = listenany(localhost, default_port)
 
+function udp_set_membership(sock::UDPSocket, group_addr::String,
+                            interface_addr::Union{Nothing, String}, operation)
+    if interface_addr === nothing
+        interface_addr = C_NULL
+    end
+    r = ccall(:uv_udp_set_membership, Cint,
+              (Ptr{Cvoid}, Cstring, Cstring, Cint),
+              sock.handle, group_addr, interface_addr, operation)
+    uv_error("uv_udp_set_membership", r)
+    return
+end
+
+"""
+    join_multicast_group(sock::UDPSocket, group_addr, interface_addr = nothing)
+
+Join a socket to a particular multicast group defined by `group_addr`.
+If `interface_addr` is given, specifies a particular interface for multi-homed
+systems.  Use `leave_multicast_group()` to disable reception of a group.
+"""
+function join_multicast_group(sock::UDPSocket, group_addr::String,
+                              interface_addr::Union{Nothing, String} = nothing)
+    return udp_set_membership(sock, group_addr, interface_addr, 1)
+end
+function join_multicast_group(sock::UDPSocket, group_addr::IPAddr,
+                              interface_addr::Union{Nothing, IPAddr} = nothing)
+    if interface_addr !== nothing
+        interface_addr = string(interface_addr)
+    end
+    return join_multicast_group(sock, string(group_addr), interface_addr)
+end
+
+"""
+    leave_multicast_group(sock::UDPSocket, group_addr, interface_addr = nothing)
+
+Remove a socket from  a particular multicast group defined by `group_addr`.
+If `interface_addr` is given, specifies a particular interface for multi-homed
+systems.  Use `join_multicast_group()` to enable reception of a group.
+"""
+function leave_multicast_group(sock::UDPSocket, group_addr::String,
+                               interface_addr::Union{Nothing, String} = nothing)
+    return udp_set_membership(sock, group_addr, interface_addr, 0)
+end
+function leave_multicast_group(sock::UDPSocket, group_addr::IPAddr,
+                               interface_addr::Union{Nothing, IPAddr} = nothing)
+    if interface_addr !== nothing
+        interface_addr = string(interface_addr)
+    end
+    return leave_multicast_group(sock, string(group_addr), interface_addr)
+end
+
 """
     getsockname(sock::Union{TCPServer, TCPSocket}) -> (IPAddr, UInt16)
 
@@ -757,32 +812,28 @@ function _sockname(sock, self=true)
     end
     iolock_end()
     uv_error("cannot obtain socket name", r)
-    if r == 0
-        port = ntoh(rport[])
-        af_inet6 = @static if Sys.iswindows() # AF_INET6 in <sys/socket.h>
-            23
-        elseif Sys.isapple()
-            30
-        elseif Sys.KERNEL ∈ (:FreeBSD, :DragonFly)
-            28
-        elseif Sys.KERNEL ∈ (:NetBSD, :OpenBSD)
-            24
-        else
-            10
-        end
-
-        if rfamily[] == 2 # AF_INET
-            addrv4 = raddress[1:4]
-            naddr = ntoh(unsafe_load(Ptr{Cuint}(pointer(addrv4)), 1))
-            addr = IPv4(naddr)
-        elseif rfamily[] == af_inet6
-            naddr = ntoh(unsafe_load(Ptr{UInt128}(pointer(raddress)), 1))
-            addr = IPv6(naddr)
-        else
-            error(string("unsupported address family: ", getindex(rfamily)))
-        end
+    port = ntoh(rport[])
+    af_inet6 = @static if Sys.iswindows() # AF_INET6 in <sys/socket.h>
+        23
+    elseif Sys.isapple()
+        30
+    elseif Sys.KERNEL ∈ (:FreeBSD, :DragonFly)
+        28
+    elseif Sys.KERNEL ∈ (:NetBSD, :OpenBSD)
+        24
     else
-        error("cannot obtain socket name")
+        10
+    end
+
+    if rfamily[] == 2 # AF_INET
+        addrv4 = raddress[1:4]
+        naddr = ntoh(unsafe_load(Ptr{Cuint}(pointer(addrv4)), 1))
+        addr = IPv4(naddr)
+    elseif rfamily[] == af_inet6
+        naddr = ntoh(unsafe_load(Ptr{UInt128}(pointer(raddress)), 1))
+        addr = IPv6(naddr)
+    else
+        error(string("unsupported address family: ", rfamily[]))
     end
     return addr, port
 end
